import flask

from pg_discuss import ext

#: Protect all views from CSRF by checking `X-Requested-With` header by
#: default.
CSRF_HEADER_CHECK_DEFAULT = True
#: Exempt a list of HTTP methods from the check (read-only methods should be
#: exempted).
CSRF_HEADER_EXEMPT_METHODS = ['GET', 'HEAD', 'OPTIONS', 'TRACE']


class CsrfHeaderExt(ext.AppExtBase):
    """Middleware to verify the request is an XHR request.
    This assumes that all, or most, data-modifying views are intended to handle
    XHR requests.

    Note that Content-Type checking alone does not necessarily protect against
    CSRF if browsers implement "HTML JSON form submission":
    http://www.w3.org/TR/html-json-forms/

    However, when the X-Requested-With header is checked, this is considered to
    be adequate protection by some:
    http://security.stackexchange.com/q/23371

    The CsrfProtectWithToken middleware (enabled by default) must be enabled
    for full token-based CSRF protection.
    """

    def init_app(self, app):
        self._app = app

        app.config.setdefault('CSRF_HEADER_CHECK_DEFAULT',
                              CSRF_HEADER_CHECK_DEFAULT)
        app.config.setdefault('CSRF_HEADER_EXEMPT_METHODS',
                              CSRF_HEADER_EXEMPT_METHODS)

        self._exempt_views = []

        if not app.config['CSRF_HEADER_CHECK_DEFAULT']:
            return

        @app.before_request
        def _csrf_protect():
            # many things come from django.middleware.csrf
            if (
                flask.request.method
                in app.config['CSRF_HEADER_EXEMPT_METHODS']
            ):
                return

            if self._exempt_views:
                if not flask.request.endpoint:
                    return

                view = app.view_functions.get(flask.request.endpoint)
                if not view:
                    return

                dest = '%s.%s' % (view.__module__, view.__name__)
                if dest in self._exempt_views:
                    return

            self.protect()

    def protect(self):
        if not flask.request.is_xhr:
            reason = (
                'XHR checking failed - X-Requested-With not set '
                'to XMLHttpRequest'
            )
            return self._error_response(reason)
        flask.request.csrf_header_valid = True

    def exempt(self, view):
        """Exclude a view from XHR checking. Takes a view function as the
        single argument.
        """
        view_location = '%s.%s' % (view.__module__, view.__name__)
        self._exempt_views.append(view_location)

    def _error_response(self, reason):
        return flask.abort(403, reason)
